/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/AGMSRAMFileSystem/RCS/mkindex.cpp,v 1.3 2003/05/22 21:03:57 agmsmith Exp $
 *
 * This program makes indices in BeOS style file systems.  It's like the
 * regular mkindex command line program but with a few options for advanced
 * index creation, such the multiple keyword indices available in
 * AGMSRAMFilesystem (ARFS).  It's also backward compatible with the old
 * mkindex utility.
 *
 * Copyright © 2003 by Alexander G. M. Smith.
 *
 * This utility program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This utility program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this utility program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * Send e-mail with suggestions for improvement to agmsmith@rogers.com.  If
 * that address doesn't work, search the web for "Alexander G. M. Smith" to
 * find my home page and contact info.
 *
 * $Log: mkindex.cpp,v $
 * Revision 1.3  2003/05/22 21:03:57  agmsmith
 * Don't fire up a BApplication.
 *
 * Revision 1.2  2003/05/14 20:25:55  agmsmith
 * Now works!
 *
 * Revision 1.1  2003/05/14 17:24:26  agmsmith
 * Initial revision
 */

/* BeOS headers. */

#include <Directory.h>
#include <fs_index.h>
#include <Volume.h>

/* Posix headers. */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* Application specific headers. */

#include "AGMSRAMFSExports.h"



/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".
 */

static void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL)
{
  char ErrorBuffer [B_PATH_NAME_LENGTH + 80 /* error message */ + 80];

  if (TitleString == NULL)
    TitleString = "Error Message:";

  if (MessageString == NULL)
  {
    if (ErrorNumber == 0)
      MessageString = "No error, no message, why bother?";
    else
      MessageString = "Something went wrong";
  }

  if (ErrorNumber != 0)
  {
    sprintf (ErrorBuffer, "%s, error code $%X/%d (%s) has occured.",
      MessageString, ErrorNumber, ErrorNumber, strerror (ErrorNumber));
    MessageString = ErrorBuffer;
  }

  fputs (TitleString, stderr);
  fputc ('\n', stderr);
  fputs (MessageString, stderr);
  fputc ('\n', stderr);
}


/******************************************************************************
 * Return a string describing a four character type code as both a number and
 * four characters.  Replace unprintable characters with a "?".
 */

static const char *TypeCodeAsString (type_code TypeCode)
{
  int         i;
  static char String [40];
  int         Letter;
  uint32      Temp32;

  Temp32 = TypeCode;
  for (i = 0; i < 4; i++)
  {
    Letter = (char) Temp32;
    Temp32 >>= 8;
    if (Letter < 32 || Letter > 127)
      Letter = '?';
    String [3 - i] = Letter;
  }

  sprintf (String + 4, " ($%X/%u)",
    (unsigned int) TypeCode, (unsigned int) TypeCode);

  return String;
}



/******************************************************************************
 * Finally, the main program which drives it all.
 */

int main (int argc, char** argv)
{
  int          ArgIndex;
  int          i;
  int          IndexFlags;
  char        *IndexNamePntr;
  type_code    IndexType;
  status_t     ReturnCode = -1;
  char        *StringPntr;
  char         TempString [256 + NAME_MAX];
  char         VolumeName [NAME_MAX];

  if (argc <= 1)
  {
    printf ("Usage: %s [-t type] [-f flags] [-m] index_name\n\n", argv[0]);
    printf ("Where index_name corresponds to the name of the attribute\n");
    printf ("you wish to index (having an index speeds up finding files\n");
    printf ("with that attribute when you use the query mechanism).  The\n");
    printf ("index will be created on the disk volume which the current\n");
    printf ("directory resides in (only indexes that volume's files).\n\n");
    printf ("type specifies the data type of the values in the index, the\n");
    printf ("old int, llong, string, float or double words still work,\n");
    printf ("plus you can now specify it as a 32 bit number (0xABCDEF12)\n");
    printf ("or as a four character code like MIMS (see TypeConstants.h\n");
    printf ("for a list).  Defaults to B_STRING_TYPE.\n\n");
    printf ("-m enables multiple keywords in the index.  So if an\n");
    printf ("attribute, like e-mail subject, is \"Make Money Fast\",\n");
    printf ("it will be broken into three words and each is added to the\n");
    printf ("index separately.  Query tests will test against each keyword\n");
    printf ("separately (any keyword that matches makes the test succeed)\n");
    printf ("so \"MAIL:subject == Money\" will find the file, or even\n");
    printf ("\"MAIL:subject==Money&&MAIL:subject==Fast&&"
      "MAIL:subject==Make\"\n");
    printf ("will find it, and all other subjects that have all three\n");
    printf ("words.  Same as B_CREATE_INDEX_MULTIPLE_KEYWORDS flag.\n\n");
    printf ("flags is a number which specifies the extra flags to pass\n");
    printf ("to the mkindex() file system operation.  OR them together to\n");
    printf ("get the desired effect.  Default is zero. Some flag values,\n");
    printf ("which may not be supported by the particular file system:\n");
    printf ("1: B_CREATE_INDEX_WITH_REBUILD\n");
    printf ("2: B_CREATE_INDEX_MULTIPLE_KEYWORDS\n");
    printf ("4: B_CREATE_INDEX_CASE_INSENSITIVE\n\n");
    printf ("Released under the GNU Lesser General Public License.\n");
    printf ("Written by Alexander G. M. Smith.  Copyright © 2003\n");
    printf ("$Revision: 1.3 $ compiled on " __DATE__ " at " __TIME__ ".\n");
    return -10;
  }

  IndexFlags = 0;
  IndexNamePntr = NULL;
  IndexType = 0;
  for (ArgIndex = 1; ArgIndex < argc; ArgIndex++)
  {
    StringPntr = argv[ArgIndex];
    if (StringPntr == NULL || StringPntr[0] == 0)
      continue;

    /* The -t type option.  Can be an old magic word, a number or a four
    character code. */

    if (strcasecmp (StringPntr, "-t") == 0)
    {
      if (ArgIndex + 1 >= argc)
        continue; // No following parameter to process.
      StringPntr = argv[ArgIndex+1];
      if (strcasecmp (StringPntr, "int") == 0)
        IndexType = B_INT32_TYPE;
      else if (strcasecmp (StringPntr, "llong") == 0)
        IndexType = B_INT64_TYPE;
      else if (strcasecmp (StringPntr, "string") == 0)
        IndexType = B_STRING_TYPE;
      else if (strcasecmp (StringPntr, "float") == 0)
        IndexType = B_FLOAT_TYPE;
      else if (strcasecmp (StringPntr, "double") == 0)
        IndexType = B_DOUBLE_TYPE;
      else if (strtoul (StringPntr, NULL, 0) != 0)
        IndexType = strtoul (StringPntr, NULL, 0);
      else if (strlen (StringPntr) == 4)
      {
        IndexType = 0;
        for (i = 0; i < 4; i++)
          IndexType = ((IndexType << 8) | (unsigned char) StringPntr[i]);
      }
      ArgIndex++; // So it doesn't think the parameter is the next argument.
    }
    else if (strcasecmp (StringPntr, "-m") == 0)
      IndexFlags |= B_CREATE_INDEX_MULTIPLE_KEYWORDS;
    else if (strcasecmp (StringPntr, "-f") == 0)
    {
      if (ArgIndex + 1 >= argc)
        continue; // No following parameter to process.
      StringPntr = argv[ArgIndex+1];
      IndexFlags |= strtoul (StringPntr, NULL, 0);
      ArgIndex++; // So it doesn't think the parameter is the next argument.
    }
    else
      IndexNamePntr = StringPntr;
  }

  /* Command line parsing finished, see if we have all the parameters. */

  if (IndexType == 0)
    IndexType = B_STRING_TYPE;

  if (IndexNamePntr == NULL)
  {
    fprintf (stderr, "%s: Index name was not specified!\n", argv[0]);
    return -21;
  }

  BDirectory CurrentDir (".");
  ReturnCode = CurrentDir.InitCheck ();
  if (ReturnCode != 0)
  {
    DisplayErrorMessage ("Unable to find current directory",
      ReturnCode, argv[0]);
    return ReturnCode;
  }

  BVolume CurrentVolume;
  CurrentDir.GetVolume (&CurrentVolume);
  ReturnCode = CurrentVolume.InitCheck ();
  if (ReturnCode != 0)
  {
    DisplayErrorMessage ("Unable to find current volume",
      ReturnCode, argv[0]);
    return ReturnCode;
  }
  CurrentVolume.GetName (VolumeName);

  /* The actual operation that this whole program condenses down to. */

  ReturnCode = fs_create_index (CurrentVolume.Device (), IndexNamePntr,
    IndexType, IndexFlags);

  if (ReturnCode != 0)
  {
    sprintf (TempString, "Unable to creating index named \"%s\", type %s, "
      "flags $%X, on volume %ld (%s)",
      IndexNamePntr, TypeCodeAsString (IndexType), IndexFlags,
      CurrentVolume.Device (), VolumeName);
    DisplayErrorMessage (TempString, ReturnCode, argv[0]);
    return ReturnCode;
  }

  printf ("%s: Successfully created index named \"%s\", type %s, flags $%X, "
    "on volume %ld (%s).\n",
    argv[0], IndexNamePntr, TypeCodeAsString (IndexType), IndexFlags,
    CurrentVolume.Device (), VolumeName);

  return ReturnCode;
}
